# Copyright 2008 Canonical Ltd.  All rights reserved.

"""Various utility functions."""

__metaclass__ = type
__all__ = [
    'BleedThroughDict',
    'camelcase_to_underscore_separated',
    'get_current_browser_request',
    'implement_from_dict',
    'make_identifier_safe',
    'safe_js_escape',
    'safe_hasattr',
    'smartquote',
    'simple_popen2',
    ]


import cgi
import re
import string
import subprocess

from simplejson import encoder

from zope.schema import getFieldsInOrder
from zope.interface import classImplements


missing = object()


class BleedThroughDict(object):
    """A stack of named dictionaries that acts as a single dictionary.

    If a key can't be found in the dictionary on top of the stack,
    this class will look in the dictionary just below the top of the
    stack, then the dictionary below that, and so on--until it
    encounters the end of the stack or a dictionary that was defined
    as 'opaque'.
    """
    missing = object()

    def __init__(self):
        """Initialize the bleed-through dictionary."""
        self.stack = []

    def push(self, name, dictionary=None, opaque=False):
        """Pushes a dictionary onto the stack.

        :arg name: The name of the dictionary to push.
        :arg dictionary: The contents of the dictionary to push.
        :arg opaque: If True, values will not 'bleed through' from
                     dictionaries lower on the stack.
        """
        if name is None:
            self.stack.append(name)
            return
        if dictionary is None:
            dictionary = {}
        self.stack.append((name, dict(dictionary), opaque))

    def pop(self):
        """Pop a tuple representing a dictionary from the stack.

        :return: A 3-tuple (name, dict, opaque). 'dict' is the
        contents of the dictionary. 'opaque' is whether or not the
        dictionary was defined as opaque. This 3-tuple can be passed
        in as the *args to append().
        """
        return self.stack.pop()

    @property
    def dict_names(self):
        """Return the names of dictionaries in the stack.

        The names are useful for passing into substack().
        """
        return [item[0] for item in self.stack if item is not None]

    @property
    def is_empty(self):
        """Is the stack empty?"""
        return len(self.stack) == 0

    def __getitem__(self, key):
        """Look up an item somewhere in the stack."""
        if self.is_empty:
            raise KeyError(key)
        value = self._get_from_top_of_stack(key)
        if value is missing:
            raise KeyError(key)
        return value

    def get(self, key, default=None):
        """Look up an item somewhere in the stack, with default fallback."""
        if self.is_empty:
            return default
        value = self._get_from_top_of_stack(key)
        if value is missing:
            return default
        return value

    def items(self):
        """Return a merged view of the items in the dictionary."""
        # Find the last opaque dictionary in the stack. We'll get our
        # values from it and every subsequent dictionary. If there are
        # no opaque dictionaries, we'll get our values from the entire
        # stack.
        dictionaries = self.stack
        for i in range(len(self.stack)-1, -1, -1):
            name, dictionary, opaque = self.stack[i]
            if opaque:
                dictionaries = self.stack[i:]
                break
        final_dictionary = {}
        for name, dictionary, opaque in dictionaries:
            final_dictionary.update(dictionary)
        return final_dictionary.items()

    def __setitem__(self, key, value):
        """Set a value in the dict at the top of the stack."""
        if self.is_empty:
            raise IndexError("Stack is empty")
        self.stack[-1][1][key] = value

    def __delitem__(self, key):
        """Delete a value from the dict at the top of the stack."""
        if self.is_empty:
            raise IndexError("Stack is empty")
        del self.stack[-1][1][key]

    def _get_from_top_of_stack(self, key):
        """Try to find the given key, starting from the top of the stack.

        If the key is not present in the dict at the top of the stack,
        this method tries the dict beneath it, and so on to the bottom.
        """
        value = missing
        for i in range(len(self.stack) - 1, -1, -1):
            name, dictionary, opaque = self.stack[i]
            value = dictionary.get(key, missing)
            if opaque or value is not missing:
                break
        return value


def implement_from_dict(class_name, interface, values, superclass=object):
    """Return a class that implements an interface's attributes.

    :param interface: The interface to implement.
    :param superclass: The superclass of the class to be generated.
    :param values: A dict of values to use when generating the class.
    :return: A class that implements 'interface'. Any attributes given
             values in 'values' will have those values in the generated
             interface. Other attributes will have their default
             values as defined in the interface. Attributes with no
             default values, will not be present.
    """
    class_dict = {}
    for name, field in getFieldsInOrder(interface):
        if field.default is not None:
            class_dict[name] = field.default
    class_dict.update(values)

    if not isinstance(superclass, tuple):
        superclass = (superclass,)
    new_class = type(class_name, superclass, class_dict)
    classImplements(new_class, interface)
    return new_class


def make_identifier_safe(name):
    """Change a string so it can be used as a Python identifier.

    Changes all characters other than letters, numbers, and underscore
    into underscore. If the first character is not a letter or
    underscore, prepends an underscore.
    """
    if name is None:
        raise ValueError("Cannot make None value identifier-safe.")
    name = re.sub("[^A-Za-z0-9_]", "_", name)
    if len(name) == 0 or name[0] not in string.letters and name[0] != '_':
        name = '_' + name
    return name


def camelcase_to_underscore_separated(name):
    """Convert 'ACamelCaseString' to 'a_camel_case_string'"""
    def prepend_underscore(match):
        return '_' + match.group(1)
    return re.sub('\B([A-Z])', prepend_underscore, name).lower()


def safe_hasattr(ob, name):
    """hasattr() that doesn't hide exceptions."""
    return getattr(ob, name, missing) is not missing


def smartquote(str):
    """Return a copy of the string, with typographical quote marks applied."""
    str = unicode(str)
    str = re.compile(u'(^| )(")([^" ])').sub(u'\\1\u201c\\3', str)
    str = re.compile(u'([^ "])(")($|[\s.,;:!?])').sub(u'\\1\u201d\\3', str)
    return str


def safe_js_escape(text):
    """Return the given text escaped for use in Javascript code.

    This will also perform a cgi.escape() on the given text.
    """
    return encoder.encode_basestring(cgi.escape(text, True))


def get_current_browser_request():
    """Return the current browser request, looked up from the interaction.

    If there is no suitable request, then return None.

    Returns only requests that provide IHTTPApplicationRequest.
    """
    from zope.security.management import queryInteraction
    from zope.publisher.interfaces.http import IHTTPApplicationRequest
    interaction = queryInteraction()
    if interaction is None:
        return None
    requests = [
        participation
        for participation in interaction.participations
        if IHTTPApplicationRequest.providedBy(participation)
        ]
    if not requests:
        return None
    assert len(requests) == 1, (
        "We expect only one IHTTPApplicationRequest in the interaction."
        " Got %s." % len(requests))
    return requests[0]


def simple_popen2(command, input, env=None, in_bufsize=1024, out_bufsize=128):
    """Run a command, give it input on its standard input, and capture its
    standard output.

    Returns the data from standard output.

    This function is needed to avoid certain deadlock situations. For example,
    if you popen2() a command, write its standard input, then read its
    standard output, this can deadlock due to the parent process blocking on
    writing to the child, while the child process is simultaneously blocking
    on writing to its parent. This function avoids that problem by using
    subprocess.Popen.communicate().
    """

    p = subprocess.Popen(
            command, env=env, stdin=subprocess.PIPE,
            stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )
    (output, nothing) = p.communicate(input)
    return output
